﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using System.Reflection;

namespace Persistence {

  #region Tokens
  public interface ISerializationToken {
  }
  public class BeginToken : ISerializationToken {
    public DataMemberAccessor Accessor;
    public int Id;
    public BeginToken(DataMemberAccessor accessor, int id) {
      this.Accessor = accessor;
      this.Id = id;
    }    
  }
  public class EndToken : ISerializationToken {
    public DataMemberAccessor Accessor;
    public int Id;
    public EndToken(DataMemberAccessor accessor, int id) {
      this.Accessor = accessor;
      this.Id = id;
    }    
  }
  public class PrimitiveToken : ISerializationToken {
    public DataMemberAccessor accessor;
    public object Data;
    public PrimitiveToken(DataMemberAccessor accessor, object data) {
      this.accessor =  accessor;
      this.Data = data;
    }    
  }
  public class ReferenceToken : ISerializationToken {
    public string Name;
    public int Id;
    public ReferenceToken(string name, int id) {
      this.Name = name;
      this.Id = id;
    }    
  }
  public class NullReferenceToken : ISerializationToken {
    public string Name;
    public NullReferenceToken(string name) {
      this.Name = name;
    }
  }
  #endregion

  #region Primitives
  public interface IPrimitiveSerializer {
    Type Type { get; }
    object Serialize(object o);
    object DeSerialize(object o);
  }
  public class String2XMLSerializer : IPrimitiveSerializer {
    public Type Type { get { return typeof(string); } }
    public object Serialize(object o) {
      return "<![CDATA[" +
        ((string)o).Replace("]]>", "]]]]><![CDATA[>") +
        "]]>";
    }
    public object DeSerialize(object o) {
      StringBuilder sb = new StringBuilder();
      foreach (string s in ((string)o).Split(
        new string[] { "<![CDATA[", "]]>" },
        StringSplitOptions.RemoveEmptyEntries)) {
        sb.Append(s);
      }
      return sb.ToString();
    }
  }
  public class Int2XMLSerializer : IPrimitiveSerializer {
    public Type Type { get { return typeof(int); } }
    public object Serialize(object o) {
      return ((int)o).ToString();
    }
    public object DeSerialize(object o) {
      return int.Parse((string)o);
    }
  }
  #endregion

  #region Generics (incomplete)
  public interface ICustomSerializer {
    bool CanSerialize(Type type);
    IEnumerable Serialize(object o);
    object DeSerialize(IEnumerable o, Type t);
  }
  public class EnumerableSerializer : ICustomSerializer {
    public bool CanSerialize(Type type) {
      if (type.GetInterface("IEnumerable") == null)
        return false;
      if (type.GetMethod("GetEnumerator", new Type[] {}) == null)
        return false;
      MethodInfo addMethod = type.GetMethod("Add");
      if (addMethod == null)
        return false;
      if (addMethod.GetParameters().Length != 1)
        return false;
      return true;
    }
    public IEnumerable Serialize(object o) {
      return (IEnumerable)o;
    }
    public object DeSerialize(IEnumerable objects, Type t) {
      object instance = Activator.CreateInstance(t);
      foreach (object o in objects) {
        t.GetMethod("Add").Invoke(instance, new object[] { o });
      }
      return instance;
    }
  }
  public class ArraySerializer : ICustomSerializer {    

    public bool CanSerialize(Type type) {
      return type.IsArray;      
    }

    public IEnumerable Serialize(object array) {
      foreach (object o in (Array)array) {
        yield return o;
      }
    }

    public object DeSerialize(IEnumerable o, Type t) {
      throw new NotImplementedException();
    }

  }
  #endregion

  public class Serializer : IEnumerable<ISerializationToken> {

    private object obj;
    private string rootName;
    private Dictionary<object, int> obj2id;
    private Dictionary<Type, IPrimitiveSerializer> primitiveSerializers;
    private List<ICustomSerializer> customSerializers;

    public Serializer(object obj, IEnumerable<IPrimitiveSerializer> primitiveSerializers) :
      this(obj, primitiveSerializers, "ROOT") { }

    public Serializer(object obj, IEnumerable<IPrimitiveSerializer> primitiveSerializers, string rootName) {
      this.obj = obj;
      this.rootName = rootName;
      this.primitiveSerializers = new Dictionary<Type,IPrimitiveSerializer>();
      foreach (IPrimitiveSerializer serializer in primitiveSerializers) {
        this.primitiveSerializers.Add(serializer.Type, serializer);
      }
      this.customSerializers = new List<ICustomSerializer>();
      customSerializers.Add(new EnumerableSerializer());
      customSerializers.Add(new ArraySerializer());
      this.obj2id = new Dictionary<object, int>();
      obj2id.Add(new object(), 0);      
    }    

    IEnumerator IEnumerable.GetEnumerator() {
      return this.GetEnumerator();
    }

    public IEnumerator<ISerializationToken> GetEnumerator() {
      DataMemberAccessor rootAccessor = new DataMemberAccessor(
        this.rootName, obj.GetType(), null, () => this.obj, null);
      IEnumerator<ISerializationToken> iterator = Serialize(rootAccessor);
      while (iterator.MoveNext())
        yield return iterator.Current;      
    }

    private IEnumerator<ISerializationToken> Serialize(DataMemberAccessor accessor) {
      object value = accessor.Get();
      if (value == null) {
        yield return new NullReferenceToken(accessor.Name);
      } else if (this.primitiveSerializers.ContainsKey(accessor.Type)) {
        yield return new PrimitiveToken(accessor, this.primitiveSerializers[accessor.Type].Serialize(value));
      } else if (this.obj2id.ContainsKey(value)) {
        yield return new ReferenceToken(accessor.Name, this.obj2id[value]);
      } else {
        int id = obj2id.Count;
        this.obj2id.Add(value, id);
        yield return new BeginToken(accessor, id);
        ICustomSerializer customSerializer = this.FindCustomSerializer(accessor.Type);
        if (customSerializer != null) {
          foreach (object obj in customSerializer.Serialize(value)) {
            IEnumerator<ISerializationToken> iterator = this.Serialize(new DataMemberAccessor(obj));
            while (iterator.MoveNext())
              yield return iterator.Current;            
          }
        } else { // composite serialization
          foreach (KeyValuePair<string, DataMemberAccessor> mapping in
            StorableAttribute.GetAutostorableAccessors(value)) {
            IEnumerator<ISerializationToken> iterator = this.Serialize(mapping.Value);
            while (iterator.MoveNext())
              yield return iterator.Current;
          }
        }
        yield return new EndToken(accessor, id);        
      }
    }

    private ICustomSerializer FindCustomSerializer(Type type) {
      foreach (ICustomSerializer s in customSerializers) {
        if (s.CanSerialize(type))
          return s;
      }
      return null;
    }                
  }

  public class XmlFormatter {

    delegate string Formatter(ISerializationToken token);

    private Dictionary<Type, Formatter> formatters;
    private int depth;

    public XmlFormatter() {
      this.formatters = new Dictionary<Type,Formatter>();
      this.formatters.Add(typeof(BeginToken), new Formatter(FormatBegin));
      this.formatters.Add(typeof(EndToken), new Formatter(FormatEnd));
      this.formatters.Add(typeof(PrimitiveToken), new Formatter(FormatData));
      this.formatters.Add(typeof(ReferenceToken), new Formatter(FormatReference));
      this.formatters.Add(typeof(NullReferenceToken), new Formatter(FormatNullReference));
      this.depth = 0;
    }
    
    public string Format(ISerializationToken token) {
      return formatters[token.GetType()](token);
    }

    private string Prefix {
      get { return new string(' ', this.depth*2); }
    }

    private string FormatBegin(ISerializationToken token) {
      BeginToken beginToken = (BeginToken)token;
      string result =
        String.Format("{0}<COMPOSITE name=\"{1}\" type=\"{2}\" id=\"{3}\">\n",
          this.Prefix, beginToken.Accessor.Name, beginToken.Accessor.Type, beginToken.Id);
      this.depth += 1;
      return result;
    }

    private string FormatEnd(ISerializationToken token) {
      EndToken endToken = (EndToken)token;
      this.depth -= 1;
      return Prefix + "</COMPOSITE>\n";
    }

    private string FormatData(ISerializationToken token) {
      PrimitiveToken dataToken = (PrimitiveToken)token;
      return String.Format("{0}<PRIMITIVE name=\"{1}\" type=\"{2}\">{3}</PRIMITIVE>\n",
        this.Prefix, dataToken.accessor.Name, dataToken.accessor.Type, dataToken.Data);      
    }

    private string FormatReference(ISerializationToken token) {
      ReferenceToken refToken = (ReferenceToken)token;
      return String.Format("{0}<REFERENCE name=\"{1}\" ref=\"{2}\"/>\n",
        this.Prefix, refToken.Name, refToken.Id);      
    }

    private string FormatNullReference(ISerializationToken token) {
      NullReferenceToken nullRefToken = (NullReferenceToken)token;
      return String.Format("{0}<NULL name=\"{1}\"/>\n",
        this.Prefix, nullRefToken.Name);
    }


  }
  
}
